class SegmentationModel:
    """A class for a generic segmentation model.

    If one needs to build a segmentation model in Keras, this class
    should provide enough attributes and properties, as well as some
    useful utilities, to make a custom and succint model.

    Attributes:
        _input_size (Tuple[int, int, int]): Size of the image, (height, width, channels)
        _filters (int): the number of output filters in the convolution
        _kernel_size (int): specifying the length of the 2D convolution window
        _dropout (float): Value for the dropout layer, between 0 and 1.
        _batch_norm (bool): Add batch normalization to every encoder block.
        _up_sample (Tuple[int, int]): Upsampling factor for rows and columns.
        _activation (str): Activation function to apply to every layer, except the last one.
        _padding (str): Type of padding to apply to the convolution layers.
        _pool (Tuple[int, int]): Pooling windows
        _l1_reg (float): Value for the L1 regularizer, applied to every convolution map
        _l2_reg (float): Value for the L2 regularizer, applied to every convolution map
        _seg_model (keras.Model): A tf.keras.Model instance.
    """
    def __init__(self):
        self._input_size = None
        self._filters = 64
        self._kernel_size = 3
        self._dropout = None
        self._batch_norm = None
        self._up_sample = (2, 2)
        self._activation = None
        self._padding = None
        self._pool = (2, 2)
        self._l1_reg = 0.0
        self._l2_reg = 0.0
        self._seg_model = None

    @property
    def model(self):
        return self._seg_model

    @property
    def filters(self):
        return self._filters

    @filters.setter
    def filters(self, value):
        self._filters = value

    @property
    def kernel_size(self):
        return self._kernel_size

    @kernel_size.setter
    def kernel_size(self, value):
        self._kernel_size = value

    @property
    def dropout(self):
        return self._dropout

    @dropout.setter
    def dropout(self, value):
        self._dropout = value

    @property
    def batch_norm(self):
        return self._batch_norm

    @batch_norm.setter
    def batch_norm(self, value):
        self._batch_norm = value

    @property
    def up_sample(self):
        return self._up_sample

    @up_sample.setter
    def up_sample(self, value):
        self._up_sample = value

    @property
    def input_size(self):
        return self._input_size

    @input_size.setter
    def input_size(self, value):
        self._input_size = value

    @property
    def activation(self):
        return self._activation

    @activation.setter
    def activation(self, value):
        self._activation = value

    @property
    def padding(self):
        return self._padding

    @padding.setter
    def padding(self, value):
        self._padding = value

    @property
    def pool(self):
        return self._pool

    @pool.setter
    def pool(self, value):
        self._pool = value

    @property
    def l1_reg(self):
        return self._l1_reg

    @l1_reg.setter
    def l1_reg(self, value):
        self._l1_reg = value

    @property
    def l2_reg(self):
        return self._l2_reg

    @l2_reg.setter
    def l2_reg(self, value):
        self._l2_reg = value

    @property
    def l1_l2_reg(self):
        return self._l1_l2_reg

    @l1_l2_reg.setter
    def l1_l2_reg(self, value):
        self._l1_l2_reg = value

    def _parse_params(self, params):
        """
        Read a dictionary of parameters and store them in the
        class attributes.

        Args:
            params (dict): A dictionary of attributes. Read the class
                docstrings to know more.
        """
        if "l1_reg" in params:
            self._l1_reg = params["l1_reg"]
        if "l2_reg" in params:
            self._l2_reg = params["l2_reg"]
        if "activation" in params:
            self._activation = params["activation"]
        if "filters" in params:
            self._filters = params["filters"]
        if "kernel_size" in params:
            self._kernel_size = params["kernel_size"]
        if "pool" in params:
            self._pool = params["pool"]
        if "padding" in params:
            self._padding = params["padding"]
        if "dropout" in params:
            self._dropout = params["dropout"]
        if "up_sample" in params:
            self._up_sample = params["up_sample"]



from tensorflow.keras.regularizers import L1L2
from .unet import simple_unet, custom_unet
from .segmodel import SegmentationModel
from .multiresunet import *


class Unet(SegmentationModel):
    """Unified interface for the simple and custom versions of the UNet.

    Attributes:
        _input_size (Tuple[int, int, int]): Size of the image, (height, width, channels)
        _filters (int): the number of output filters in the convolution
        _kernel_size (int): specifying the length of the 2D convolution window
        _dropout (float): Value for the dropout layer, between 0 and 1.
        _batch_norm (bool): Add batch normalization to every encoder block.
        _up_sample (Tuple[int, int]): Upsampling factor for rows and columns.
        _activation (str): Activation function to apply to every layer, except the last one.
        _padding (str): Type of padding to apply to the convolution layers.
        _pool (Tuple[int, int]): Pooling windows
        _l1_reg (float): Value for the L1 regularizer, applied to every convolution map
        _l2_reg (float): Value for the L2 regularizer, applied to every convolution map
        _seg_model (keras.Model): A tf.keras.Model instance.
        _from_keras (dict): Additional parameters from the Keras API than can apply ONLY to
            convolutional layers.
    """

    def __init__(
        self,
        input_size=(512, 512, 3),
        variant=None,
        parameters=None,
        keras_parameters=None,
    ):
        """Creates a Unet model with the specified variant.

        For the custom version, almost ALL parameters must be specified.

        Args:
            input_size (Tuple[int, int, int]): Size of the image, (height, width, channels)
            variant (str): Either "simple" or "custom", specifying the type of model
            parameters (dict): All the parameters to add from the SegmentationModel specification.
                See the documentation for more on the attributes.
            keras_parameters (dict): Additional parameters from the Keras API than can apply ONLY to
                convolutional layers.
        """
        super().__init__()
        self._input_size = input_size
        self._from_keras = keras_parameters
        self._variant = variant

        # The simple model has a ReLU activation and "same" padding
        if self._variant is "simple":
            self._activation = "relu"
            self._padding = "same"
        # For the custom model, parse all the parameters from the dictionary
        if self._variant is "custom":
            if parameters is None:
                raise ValueError("For a custom network, parameters must be set.")
            # Create the model with the specified parameters
            self._parse_params(parameters)

    def _simple_init(self):
        # Build the SegmentationModel API dictionary of parameters
        options = {"activation": self._activation, "padding": self._padding}
        self._seg_model = simple_unet(self._input_size, options)

    def _custom_init(self):
        # Build the Keras API dictionary of parameters
        conv = {
            "filters": self._filters,
            "kernel_size": self._kernel_size,
            "activation": self._activation,
            "padding": self._padding,
            "kernel_regularizer": L1L2(l1=self._l1_reg, l2=self._l2_reg),
        }
        # Add additional Keras parameters if passed
        if self._from_keras is not None:
            conv = {**conv, **self._from_keras}
        # Build the SegmentationModel API dictionary of parameters
        options = {
            "pool": self._pool,
            "dropout": self._dropout,
            "batch_norm": self._batch_norm,
            "up_sample": self._up_sample,
        }
        self._seg_model = custom_unet(self._input_size, conv, **options)

    def collect(self):
        """Gather all the information from the model, inside and outside.

        Generate a Keras instance from all the information.
        This way the model can be instantiated
        from the beginning or by updating the attributes individually until you are
        satisfied with the model you have.

        Returns:
            _seg_model (keras.Model): A tf.keras.Model instance with all the information
                from the attributes.
        """
        if self._variant is "simple":
            self._simple_init()

        if self._variant is "custom":
            self._custom_init()

        return self._seg_model



import tensorflow as tf


def jaccard_index(y_true, y_pred):
    """Evaluate the Jaccard index.

    Assuming that `y_true` and `y_pred` are images,
    the Jaccard index evaluates segmentation maps and its effectiveness.
    If the value is one, the segmentation map predicted is exact.

    Args:
        y_true: The ground truth.
        y_pred: The predicted value.

    Returns:
        result: Scalar that determines the intersection over union value.
    """
    y_true_f = tf.reshape(y_true, shape=[-1])
    y_pred_f = tf.reshape(y_pred, shape=[-1])
    intersection = tf.reduce_sum(y_true_f * y_pred_f)
    union = tf.reduce_sum(y_true_f) + tf.reduce_sum(y_pred_f)
    result = (intersection + 1.0) / (union - intersection + 1.0)
    result = tf.reduce_mean(result)
    result = tf.cast(result, tf.float32)

    return result


def dice_coef(y_true, y_pred, smooth=1.0):
    """Evaluate the Sorensen-Dice coefficient.

    Assuming that `y_true` and `y_pred` are images.
    The Sorensen-Dice coefficient is a close relative of the Jaccard index, and it should
    almost always be logged simultaneously.

    Args:
        y_true: The ground truth.
        y_pred: The predicted value.

    Returns:
        result: Scalar that determines the segmentation error.
    """

    y_true_f = tf.reshape(y_true, shape=[-1])
    y_pred_f = tf.reshape(y_pred, shape=[-1])
    intersection = tf.reduce_sum(y_true_f * y_pred_f)
    numerator = (2.0 * intersection) + smooth
    denom = tf.reduce_sum(y_true_f) + tf.reduce_sum(y_pred_f) + smooth
    result = numerator / denom

    return result


def _static_binarization(x):
    """Take a TensorFlow Tensor object and statically binarize it.
    
    This function makes the assumption that `x`
     has values between 0 and 1. If the element value of `x` is >= 0.5,
    then an integer 1 is assigned, otherwise, an integer 0 is assigned.

    Args:
        x: Tensor to binarize

    Returns:
        new_x: Binarized tensor
    """
    new_x = tf.where(x >= 0.5, 1, 0)
    new_x = tf.cast(new_x, tf.float32)

    return new_x


def _up_dp_qp(x, y):
    """Obtain under, over and overall error segmentation rates.

    By binarizing a ground truth and predicted segmentation maps,
    this function obtains:
    - Qp, the number of pixels that should
        be included in the segmentation result but are not
    - Up, the number of pixels that should be
        excluded from the segmentation result but actually included
    - Dp, the number of pixels that should be included in the segmentation result
        and are also actually included.
    This implementation leverages logical operations and invertions for
    a faster computation of the values.

    Args:
        x: The ground truth
        y: The predicted value

    Returns:
        u_p: the calculated value for Up
        d_p: the calculated value for Dp
        q_p: the calculated value for Qp
    """
    y_t = tf.reshape(x, shape=[-1])
    y_p = tf.reshape(y, shape=[-1])
    y_t = _static_binarization(y_t)
    y_p = _static_binarization(y_p)
    d_p = tf.reduce_sum(y_t * y_p)
    q_p = tf.reduce_sum(y_t * (1.0 - y_p))
    u_p = tf.reduce_sum(y_p * (1.0 - y_t))

    return u_p, d_p, q_p


def o_rate(y_true, y_pred):
    """Calculate the over segmentation error.
    
    Assuming `y_true` and `y_pred` are images, this function calculates
    the overall segmentation error, OR, defined as:

    Args:
        y_true: with the ground truth.
        y_pred: with the predicted value.

    Returns:
        result: Constant Tensor with the OR value.
    """
    u_p, d_p, q_p = _up_dp_qp(y_true, y_pred)
    # Add a smoothing factor of 1.0 to prevent division by zero
    result = q_p / (u_p + d_p + 1.0)
    # Use the number of images as a means of normalizing the result
    num_data = tf.cast(tf.shape(y_true), result.dtype)
    result = tf.cast(result / num_data[0], tf.float32)

    return result


def u_rate(y_true, y_pred):
    """Calculate the under segmentation error.
    
    Assuming `y_true` and `y_pred` are images, this function calculates
    the overall segmentation error, UR, defined as:

    UR = Up / (Up + Dp)

    Args:
        y_true: with the ground truth.
        y_pred: with the predicted value.

    Returns:
        result: Constant Tensor with the UR value.
    """
    u_p, d_p, _ = _up_dp_qp(y_true, y_pred)
    # Add a smoothing factor of 1.0 to prevent division by zero
    result = u_p / (u_p + d_p + 1.0)
    # Use the number of images as a means of normalizing the result
    num_data = tf.cast(tf.shape(y_true), result.dtype)
    result = tf.cast(result / num_data[0], tf.float32)

    return result


def err_rate(y_true, y_pred):
    """Calculate the overall segmentation error.
    
    Assuming `y_true` and `y_pred` are images, this function calculates
    the overall segmentation error, ER, defined as:

    ER = (Qp + Up) / Dp

    Args:
        y_true: with the ground truth.
        y_pred: with the predicted value.

    Returns:
        result: Constant Tensor with the ER value.
    """
    u_p, d_p, q_p = _up_dp_qp(y_true, y_pred)
    # Add a smoothing factor of 1.0 to prevent division by zero
    result = (q_p + u_p) / (d_p + 1.0)
    # Use the number of images as a means of normalizing the result
    num_data = tf.cast(tf.shape(y_true), result.dtype)
    result = tf.cast(result / num_data[0], tf.float32)
    result = tf.math.abs(result)

    return result




import tensorflow as tf
from .metrics import jaccard_index

def ternaus_loss(y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
    """Evaluate the Ternaus loss.

    Loss inspired by TernausNet
    https://arxiv.org/abs/1801.05746
    A (hopefully) smooth and differentiable combination between binary cross-entropy
    and the Jaccard index for better segmentation training.

    Arguments:
        y_true: The ground truth.
        y_pred: The predicted value.

    Returns:
        loss: The segmentation error.
    """
    loss = tf.keras.losses.binary_crossentropy(y_true, y_pred) - tf.math.log(
        jaccard_index(y_true, y_pred)
    )

    return loss


from segmed.models import unet
from segmed.metrics import metrics as mts
import tensorflow as tf


def train_segmed(
    model,
    img_path,
    mask_path,
    batch_size=16,
    epochs=25,
    steps_per_epoch=3125,
    val_split=0.2,
    optimizer=tf.keras.optimizers.Adam(),
    monitor="val_jaccard_index",
    model_file="unet_simple.h5",
    seed=1,
    show=False,
):

    """A simple utility function for training the any `segmed` model.

    Takes two paths for images and segmentation maps and rescales them, splits
    them into training and validation, while saving the best model trained.
    The paths must comply to the Keras API convention as specified in their
    documentation.

    Args:
        model (tf.keras.Model): A keras model instance.
        img_path (str): The relative path were the images are located.
        mask_path (str): The relative path were the maps are located.
        batch_size (int): Size of the batch to be processed.
        epochs (int): Number of epochs to train the model.
        steps_per_epoch (int): Total number of steps (batches of samples) to yield 
            before declaring one epoch finished and starting the next epoch.
        val_split (float): Value between 0.0 and 1.0 representing the size in percentage
            to split the dataset.
        optimizer (`Optimizer`): A Keras optimizer instance with a valid syntax from TensorFlow.
        monitor (str): Quantity to monitor during training; follow the Keras convention.
        model_file (str): File to create for the ModelCheckpoint callback from Keras.
        seed (int): Value to seed the ImageDataGenerator and always retrieve the same batch
            of pairs of images.
        show (bool): If true, plots a pair of size `batch_size` to see if the image and its
            segmentation maps are consistent. For debugging purposes only.

    Returns:
        history (`History`): A `History` object. Its `History.history` attribute is a record of
            training loss values and metrics values at successive epochs, as well as validation
            loss values and validation metrics values (if applicable).
    """

    # Rescale and convert to float32 both subsets
    data_gen_args = {
        "rescale": 1.0 / 255.0,
        "validation_split": val_split,
        "dtype": tf.float32,
    }

    # Crea the training generators with the defined transformations
    image_datagen = tf.keras.preprocessing.image.ImageDataGenerator(**data_gen_args)
    mask_datagen = tf.keras.preprocessing.image.ImageDataGenerator(**data_gen_args)
    # Take images from directories
    image_generator_train = image_datagen.flow_from_directory(
        img_path,
        class_mode=None,
        color_mode="rgb",
        batch_size=batch_size,
        seed=seed,
        subset="training",
    )
    mask_generator_train = mask_datagen.flow_from_directory(
        mask_path,
        class_mode=None,
        color_mode="grayscale",
        batch_size=batch_size,
        seed=seed,
        subset="training",
    )
    # Combine both generators
    # This need to be a generator of generators, see the following
    # https://github.com/tensorflow/tensorflow/issues/32357
    train_generator = (
        pair for pair in zip(image_generator_train, mask_generator_train)
    )

    # And now the validation generators
    image_generator_val = image_datagen.flow_from_directory(
        img_path,
        class_mode=None,
        color_mode="rgb",
        batch_size=batch_size,
        seed=seed,
        subset="validation",
    )
    mask_generator_val = mask_datagen.flow_from_directory(
        mask_path,
        class_mode=None,
        color_mode="grayscale",
        batch_size=batch_size,
        seed=seed,
        subset="validation",
    )
    # Combine both generators, with same issue as before
    val_generator = (pair for pair in zip(image_generator_val, mask_generator_val))

    # DEBUGGING ONLY, for checking that both image and maps are batched together
    if show:
        import matplotlib.pyplot as plt

        for i, j in train_generator:
            plt.figure(0)
            plt.imshow(i[0, ...])
            plt.figure(1)
            plt.imshow(j[0, ..., 0])
            plt.show()

    # Define the checkpoint callback, always maximum mode for custom metrics
    checkpoint = tf.keras.callbacks.ModelCheckpoint(
        model_file, monitor=monitor, verbose=1, save_best_only=True, mode="max"
    )

    # Compile the model with an and custom metrics
    model.compile(
        loss=tf.keras.losses.binary_crossentropy,
        optimizer=optimizer,
        metrics=[mts.jaccard_index, mts.dice_coef],
    )

    # Create history of model and return it
    history = model.fit_generator(
        train_generator,
        steps_per_epoch=steps_per_epoch,
        epochs=epochs,
        callbacks=[checkpoint],
        verbose=1,
        validation_data=val_generator,
        validation_steps=steps_per_epoch,
    )

    return history



from skimage import io
from sklearn.feature_extraction import image
import numpy as np
from tensorflow.keras.preprocessing.image import ImageDataGenerator


def extract_data(train_path, label_path=None, rgb=False):
    """Load images to memory from directories.

    Extracts data from train_path and label_path and returns a normalized
    copy of the information. These are expected to be images, and the normalization
    is just converting pixel values to a 0-1 range.

    Args:
        train_path (str): String with the path for the full images.
        label_path (str): String with the path for the segmentation maps.
        rgb (bool): Boolean value to determine if the images are color or grayscale images.

    Returns:
        X_train, y_train (ndarray): Arrays of the images and their segmentation maps, normalized.
        X_train (ndarray): Array of the images and their segmentation maps, normalized; when there
            is no `label_path`.
    """

    # Import images as a collection
    X_train = io.ImageCollection(train_path).concatenate()
    
    # Reshape the array in case it is needed
    if rgb:
        # First check if for whatever reason the array is not RGB already
        if not X_train.shape[-1] == 3:
            X_train = X_train[:, :, :, 3]
    else:
        # If not RGB, just reshape to having just one channel, grayscale
        X_train = X_train[:, :, :, np.newaxis]
    
    # Always convert to a valid type and normalize
    X_train = X_train.astype("float32")
    X_train /= 255.0

    # Do the same to the segmentation maps, if passed
    if not label_path is None:
        y_train = io.ImageCollection(label_path).concatenate()
        # The segmentation maps should always be shape (:, :, :, 1)
        y_train = y_train[:, :, :, np.newaxis]
        # Convert to a valid type and normalize
        y_train = y_train.astype("float32")
        y_train /= 255.0

        return X_train, y_train

    return X_train


def split_images(x, y=None, size=(128, 128), num_part=4):
    """Split images into smaller images.

    Takes two arrays of images, x,y, and splits them into num_part number
    of random patches.

    Args:
        x: Numpy ndarray with images.
        y: Numpy ndarray with images.
        size: Tuple with two integer values, (height, width) of the resulting patches.
        num_part: Integer value; the number of resulting patches.

    Returns:
        x_imgs, y_imgs: Numpy ndarrays with the patches of the original images.
    """
    x_patches = image.PatchExtractor(patch_size=size, max_patches=num_part, random_state=0)
    x_imgs = x_patches.transform(x)
    # Check if number of channels is the same for grayscale
    if x.shape[-1] != x_imgs.shape[-1]:
        x_imgs = x_imgs[:, :, :, np.newaxis]

    if not y is None:
        y_patches = image.PatchExtractor(patch_size=size, max_patches=num_part, random_state=0)
        y_imgs = y_patches.transform(y)

        # Check if number of channels is the same for grayscale
        if y.shape[-1] != y_imgs.shape[-1]:
            y_imgs = y_imgs[:, :, :, np.newaxis]

        return x_imgs, y_imgs

    return x_imgs

def image_mask_augmentation(x, y, batch_size=4, transformations=None, seed=6):
    """Use Keras to transform images in several ways.

    Takes two sets of images and transforms it using ImageDataGenerator from Keras.
    For more information about the possible transformation, visit the official
    Keras documentation.

    Args:
        x: Numpy ndarray with images.
        y: Numpy ndarray with images.
        batch_size: Integer value for the number of images to take simultaneously.
        transformations: Dictionary with the specified transformations to perform, using Keras API.
        seed: Integer value left fixed for reproducibility purposes.

    Returns:
        generator: A Python generator that yields both datasets transformed.
    """
    # Always perform some basic transformations
    if transformations is None:
        transformations = dict(
        rotation_range=10.0,
        height_shift_range=0.02,
        shear_range=5,
        horizontal_flip=True,
        vertical_flip=False,
        fill_mode="constant"
    )

    datagen_x = ImageDataGenerator(**transformations)
    datagen_x.fit(x, augment=True, seed=seed)
    datagen_y = ImageDataGenerator(**transformations)
    datagen_y.fit(y, augment=True, seed=seed)

    x_aug = datagen_x.flow(x, batch_size=batch_size, seed=seed)
    y_aug = datagen_y.flow(y, batch_size=batch_size, seed=seed)

    generator = zip(x_aug, y_aug)

    return generator



import os
import json
import datetime

from functools import wraps
from time import time
from typing import Callable, Any, Optional

# TODO: Refactor docstrings to be more specific of functionality
def time_this(f: Callable) -> Callable:
    """
      Code snippet taken from : https://stackoverflow.com/questions/1622943/timeit-versus-timing-decorator
      Modified to work on Python 3.7
      This simply wraps the function and takes the time it takes to compute on a single run. 
      
      One could log it and then perform statistical analysis, rather than using timeit which 
      disables the garbage collector.
    """

    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        exec_time = te - ts
        print(f"func:{f.__name__} args:[{args}, {kw}] took: {exec_time:.4f} s")
        return result

    return wrap


def time_log(path_to_logfile: Optional[str] = None) -> Callable:
    """Logs the time to compute a function.

    Logging is done using the json-lines format : http://jsonlines.org/
    
    Each log consists of:
    {
    "datetimeUTC": (Standard Greenwich time at function call),
    "function": (Name of the function that was called, given by function.__name__),
        "args": (Values of positional arguments, if JSON-serializable
                str(type(arg)) for each non-JSON-serializable argument),
        "kwargs": (Idem, for keyword arguments),
        "time": (Time required to execute the wrapped function, calculated as follows:
                ts = time()
                result = f(*args, **kw)
                te = time()
                exec_time = te - ts
                )
    }
    
    Args:
        path_to_logfile: string containing a valid path to a logfile.
            If no path is specified, it will generate a default path :
            
            "`pwd`/time_logs.jsonl"
            pwd is obtained using Python's os module, i.e. os.path.realpath('.')
    
    Returns:
        timed: another decorator which will measure time needed to execute code.
    """
    if not path_to_logfile:
        path_to_logfile = os.path.join(os.path.realpath("."), "time_logs.jsonl")

    def timed(f: Callable):
        @wraps(f)
        def wrap(*args, **kw):
            ts = time()
            result = f(*args, **kw)
            te = time()
            exec_time = te - ts
            json_cast = lambda x: x if is_jsonable(x) else str(type(x))
            data = {
                "datetimeUTC": str(datetime.datetime.utcnow()),
                "function": f.__name__,
                "args": [json_cast(arg) for arg in args],
                "kwargs": {key: json_cast(kw[key]) for key in kw.keys()},
                "time": exec_time,
            }
            with open(path_to_logfile, "a") as log:
                log.write(json.dumps(data) + "\n")
            return result

        return wrap

    return timed


def is_jsonable(x: Any) -> bool:
    """Verify if object is JSON-serializable.
        
    Args:
        x : Literally any Python object.
                
    Returns:
        True : If json.dumps(x) succeeds.
        False : If json.dumps(x) raises any kind of Exception.
    """
    is_arg_jsonable = None
    try:
        json.dumps(x)
        is_arg_jsonable = True
        return is_arg_jsonable
    except:
        is_arg_jsonable = False
        return is_arg_jsonable



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['segmed', 'segmed.metrics', 'segmed.models', 'segmed.train', 'segmed.utils']

package_data = \
{'': ['*']}

install_requires = \
['scikit-image>=0.16.2,<0.17.0',
 'scikit-learn>=0.21.3,<0.22.0',
 'tensorflow>=2.0,<3.0']

setup_kwargs = {
    'name': 'segmed',
    'version': '0.6.12',
    'description': 'Applying Deep Learning to medical image segmentation tasks',
    'long_description': '# SegMed [![Build Status](https://travis-ci.org/DCI-NET/segmed.svg?branch=master)](https://travis-ci.org/DCI-NET/segmed)\n\nThis is a collection of Deep Learning semantic segmentation models to use for\nspecific tasks, namely medical images, cells, histological data and related.\n\n## Models\n\nThe models here presented are just two, namely the [U-Net](https://arxiv.org/pdf/1505.04597.pdf)\nand the [MultiResUNet](https://arxiv.org/pdf/1902.04049.pdf). These models have been a very good\napplication of Fully Convolutional Networks to the medical image segmentation task, and are very\nwell suited for it.\n\n## Implementation\n\nEverything is implemented with [TensorFlow 2.0](tensorflow.org), using the newly acquired Keras API\nwithin TensorFlow. This allows for the flexibility and completeness of using the full TensorFlow\nlibrary while still having very good scripting capabilities with Keras.\n\n## Dependencies\n\nThere are several ways to install this package.\n\n### Using `pip`\n\nThe easiest way to install this package is using `pip` with the following command\n\n    pip install segmed\n\nalthough it is _highly encouraged_ to do this inside a virtual environment.\n\nIf using [Colaboratory](https://colab.research.google.com/notebooks/welcome.ipynb)\nthen this is the **preferred** way to use the package.\n\n**IMPORTANT**: When using Colaboratory one must always install `TensorFlow 2.0` first and then install `segmed`,\ni.e. this package, using the following commands in a cell within a Colaboratory notebook:\n```\n!pip install --upgrade pip\n!pip install segmed\n!pip install --upgrade tensorflow\n```\n\n### Using `poetry`\n\n[`poetry`](https://poetry.eustace.io/) is supported, by following the\n[installation](https://poetry.eustace.io/docs/#installation) instructions to get `poetry` installed, the following\ncommand should install `segmed` in a virtual environment:\n```shell\n# clone the repository\ngit clone https://github.com/DCI-NET/segmed\n# run poetry and install\npoetry install\n```\n\n`poetry` is a next-gen dependecy manager and makes everything a lot easier.\n\n### Using `conda`\n\nThis package also comes with a\n[conda environment](https://docs.conda.io/projects/conda/en/latest/user-guide/tasks/manage-environments.html)\nfor those that use the [Anaconda distribution](https://www.anaconda.com/distribution/).\n\nTo install the environment make sure you have [conda](https://conda.io/en/latest/) installed, then run the following\n\n    conda env create -f segmed_env.yml\n\nthis should ask you to confirm the installation, say yes and proceed with the installation. After that, activate the newly\ncreated environment\n\n    conda activate segmed\n\nand now you are ready to run the code within this repository.\n\n## Unit tests\n\nThis repository has some unit tests available that should be running constantly in the background,\nand the status of the current code build is displayed in the badge above (the one right to the title).\n\nOne can manually run the tests, too. You can download this repository with `git` like so:\n\n    git clone https://github.com/DCI-NET/segmed.git\n\nThen, you install [pytest](https://pytest.org/en/latest/) and just run the following command\n\n    pytest\n\nand the test suite should start running, with a few import and API warnings, but everything should pass\nif the badge above says _passing_ in green.\n\n## Examples\n\nThis repository also has some (_very_) barebones examples of how to use these models.\nHowever, they were run in a local machine and most of the data cannot be used.\nThese examples should be used as a _tutorial_ for the package,\njust to have a basic idea of how to run and create an image segmentation pipeline with `segmed`, \nbut you will _not_ be able to rerun the notebooks.\n\nThe reason for this is that most of the datasets are **very** large, so they cannot be bundled\nwith this repository; some other datasets **cannot** be redistributed as per request of the original authors.\n\nEither way, **all** of the trained models and weights are **freely available** upon request.\n\n### Demo\n\nFor completeness, here is a simple example. Assuming you have followed the instructions and everything is installed\ncorrectly, you can do the following to train a simple U-Net model:\n```python\nfrom segmed.train import train_unet\nfrom segmed.models import Unet\n\n# Define some example hyperparameters\nbatch_size = 8\nepochs = 50\nsteps_per_epoch=100\n\n# Declare the paths to use (following the Keras convention)\n# https://www.tensorflow.org/versions/r2.0/api_docs/python/tf/keras/Model#fit_generator\ndata_path = "general/path/to/images"\nimg_path = data_path + "augmented/images/path"\nmasks_path = data_path + "augmented/masks/path"\nmodel_file = "path/to/save/model/unet_model.h5"\n\n# Create a Unet (custom) model with a regularizer and\n# batch normalization\ncustom_params = {\n    "activation": "relu",\n    "padding": "same",\n    "batch_norm": True,\n    "l2_reg": 0.995\n}\nmodel = Unet(variant="custom", parameters=custom_params)\n# Train the model!\nhistory = train_unet(\n    model,\n    img_path,\n    masks_path,\n    batch_size=batch_size,\n    epochs=epochs,\n    steps_per_epoch=steps_per_epoch,\n    model_file=model_file,\n)\n```\nNow, the training should have started and you\'re good to go!\n\n## Datasets\n\nThe datasets employed in some of the parts of this repository are the following:\n\n- [ISBI 2012 Challenge](http://brainiac2.mit.edu/isbi_challenge/home)\n- [Motion-based Segmentation and Recognition Dataset](http://mi.eng.cam.ac.uk/research/projects/VideoRec/CamVid/)\n- [DRIVE: Digital Retinal Images for Vessel Extraction](https://www.isi.uu.nl/Research/Databases/DRIVE/)\n- [Fast and Robust Segmentation of White Blood Cell Images by Self-supervised Learning](https://github.com/zxaoyou/segmentation_WBC)',
    'author': 'Edwin Bedolla',
    'author_email': 'developeredwin@gmail.com',
    'url': 'https://github.com/DCI-NET/segmed',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from tensorflow import keras as K


def conv2d(x, filters, shape, padding="same", strides=(1, 1), activation="relu"):
    """
    2D Convolutional layers with Batch Normalization
    
    Args:
        x: Keras layer, the input to the feature map
        filters: Int representing the number of filters to use
        shape: Tuple with two integer values (number of rows, number of columns)
        padding: String that determines the padding mode
        strides: Tuple of two integer values that represent the strides
        activation: String that defines the activation function
    
    Returns:
        x: A Keras layer
    """

    x = K.layers.Conv2D(
        filters, shape, strides=strides, padding=padding, use_bias=False
    )(x)
    x = K.layers.BatchNormalization(scale=False)(x)

    if activation is None:
        return x

    x = K.layers.Activation(activation)(x)

    return x


def MultiResBlock(u_val, input, alpha=1.67):
    """MultiRes Block, as defined in the paper.

    Alpha is a constant value that controls
    the number of parameters in the block.
    
    Args:
        U: Integer value for the number of filters.
        input: A Keras layer.
    
    Returns:
        out: A Keras layer.
    """
    # Calculate the value of W as defined in the paper.
    weight = u_val * alpha
    # The first 1x1 map, to preserve dimensions
    dimension_conservation = conv2d(
        input,
        int(weight * 0.167) + int(weight * 0.333) + int(weight * 0.5),
        (1, 1),
        activation=None,
        padding="same",
    )
    # First 3x3 map, adjusted with W / 6
    conv3x3 = conv2d(
        input, int(weight * 0.167), (3, 3), activation="relu", padding="same"
    )
    # Second 3x3 map, adjusted with W / 3
    conv5x5 = conv2d(
        conv3x3, int(weight * 0.333), (3, 3), activation="relu", padding="same"
    )
    # Third 3x3 map, adjusted with W / 2
    conv7x7 = conv2d(
        conv5x5, int(weight * 0.5), (3, 3), activation="relu", padding="same"
    )
    # Concatenate all three 3x3 maps
    out = K.layers.Concatenate()([conv3x3, conv5x5, conv7x7])
    out = K.layers.BatchNormalization()(out)
    # And add the new 7x7 map with the 1x1 map, batch normalized
    out = K.layers.add([dimension_conservation, out])
    out = K.layers.Activation("relu")(out)
    out = K.layers.BatchNormalization()(out)

    return out


def ResPath(filters, input, length=None):
    """ResPath, to mitigate the semantic gap in the architecture.

    This function creates a path with just one combination of residual
    and feature maps, and this can easily be extended with the length
    argument.
    
    Args:
        filters: Integer value corresponding to the number of filters.
        length: Integer value with the length of the path, number of maps.
        input: Keras layer.
    
    Returns:
        out: Keras layer.
    """
    # First residual connection
    residual = conv2d(input, filters, (1, 1), activation=None, padding="same")
    # And first feature map
    out = conv2d(input, filters, (3, 3), activation="relu", padding="same")
    # Add the layers and batch normalize
    out = K.layers.add([residual, out])
    out = K.layers.Activation("relu")(out)
    out = K.layers.BatchNormalization()(out)
    # If there is more maps to add, we add them with this loop
    if not length is None:
        for _ in range(length - 1):

            residual = out
            residual = conv2d(
                residual, filters, (1, 1), activation=None, padding="same"
            )

            out = conv2d(out, filters, (3, 3), activation="relu", padding="same")

            out = K.layers.add([residual, out])
            out = K.layers.Activation("relu")(out)
            out = K.layers.BatchNormalization()(out)

    return out


def MultiResUnet(input_size=(256, 256, 3)):
    """The MultiResUNet neural network.

    A TensorFlow implementation of the MultiResUNet architecture as defined in the
    following paper:
        https://arxiv.org/abs/1902.04049
    
    This is a variant of the U-Net, with additional blocks and paths to help mitigate
    semantic gaps and to obtain better characteristics from the images and maps.
    
    Args:
        input_size: Tuple of three integers (height, width, number of channels) that
            describe the input images.
    
    Returns:
        model: A Keras model instance.
    """

    inputs = K.layers.Input((input_size))

    mresblock_1 = MultiResBlock(32, inputs)
    pool_1 = K.layers.MaxPooling2D(pool_size=(2, 2))(mresblock_1)
    mresblock_1 = ResPath(32, mresblock_1, 4)

    mresblock_2 = MultiResBlock(64, pool_1)
    pool_2 = K.layers.MaxPooling2D(pool_size=(2, 2))(mresblock_2)
    mresblock_2 = ResPath(64, mresblock_2, 3)

    mresblock_3 = MultiResBlock(128, pool_2)
    pool_3 = K.layers.MaxPooling2D(pool_size=(2, 2))(mresblock_3)
    mresblock_3 = ResPath(128, mresblock_3, 2)

    mresblock_4 = MultiResBlock(256, pool_3)
    pool_4 = K.layers.MaxPooling2D(pool_size=(2, 2))(mresblock_4)
    mresblock_4 = ResPath(256, mresblock_4)

    mresblock5 = MultiResBlock(512, pool_4)

    up_6 = K.layers.Conv2DTranspose(256, (2, 2), strides=(2, 2), padding="same")(
        mresblock5
    )
    up_6 = K.layers.Concatenate()([up_6, mresblock_4])
    mresblock_6 = MultiResBlock(256, up_6)

    up_7 = K.layers.Conv2DTranspose(128, (2, 2), strides=(2, 2), padding="same")(
        mresblock_6
    )
    up_7 = K.layers.Concatenate()([up_7, mresblock_3])
    mresblock7 = MultiResBlock(128, up_7)

    up_8 = K.layers.Conv2DTranspose(64, (2, 2), strides=(2, 2), padding="same")(
        mresblock7
    )
    up_8 = K.layers.Concatenate()([up_8, mresblock_2])
    mresblock8 = MultiResBlock(64, up_8)

    up_9 = K.layers.Conv2DTranspose(32, (2, 2), strides=(2, 2), padding="same")(
        mresblock8
    )
    up_9 = K.layers.Concatenate()([up_9, mresblock_1])
    mresblock9 = MultiResBlock(32, up_9)

    conv_10 = conv2d(mresblock9, 1, (1, 1), activation="sigmoid")

    model = K.models.Model(inputs=[inputs], outputs=[conv_10])

    return model




from tensorflow import keras as K
from warnings import warn


def simple_unet(input_size, conv):
    """
    Implementation of the U-Net model, using Concatenation instead of
    crop and place for the semantic gap.
    
    Args:
        input_size (Tuple[int, int, int]): (height, width, channels), i.e.
        The information of the image.
        Must always be a multiple of 32, e.g. 256, 512.
        conv (dict): Keyword parameters from the Keras specification. This dictionary
            will affect ALL convolutional layers in the network.

    Returns:
        model: A tf.keras.Model instance.
    """

    # Take in the inputs
    inputs = K.layers.Input(input_size)

    # First encoder block
    conv_1 = K.layers.Conv2D(64, 3, **conv)(inputs)
    conv_1 = K.layers.Conv2D(64, 3, **conv)(conv_1)
    pool_1 = K.layers.MaxPooling2D(pool_size=(2, 2))(conv_1)

    # Second encoder block
    conv_2 = K.layers.Conv2D(128, 3, **conv)(pool_1)
    conv_2 = K.layers.Conv2D(128, 3, **conv)(conv_2)
    pool_2 = K.layers.MaxPooling2D(pool_size=(2, 2))(conv_2)

    # Third encoder block
    conv_3 = K.layers.Conv2D(256, 3, **conv)(pool_2)
    conv_3 = K.layers.Conv2D(256, 3, **conv)(conv_3)
    pool_3 = K.layers.MaxPooling2D(pool_size=(2, 2))(conv_3)

    # Fourth encoder block
    conv_4 = K.layers.Conv2D(512, 3, **conv)(pool_3)
    conv_4 = K.layers.Conv2D(512, 3, **conv)(conv_4)
    # drop_4 = K.layers.Dropout(0.5)(conv_4)
    pool_4 = K.layers.MaxPooling2D(pool_size=(2, 2))(conv_4)

    # Encoder-decoder conection
    conv_5 = K.layers.Conv2D(1024, 3, **conv)(pool_4)
    conv_5 = K.layers.Conv2D(1024, 3, **conv)(conv_5)
    # drop_5 = K.layers.Dropout(0.5)(conv_5)

    # First decoder block
    up_6 = K.layers.UpSampling2D(size=(2, 2))(conv_5)
    up_6 = K.layers.Conv2D(512, 2, **conv)(up_6)

    # Concatenation of first decoder and fourth encoder blocks
    merge_6 = K.layers.Concatenate()([conv_4, up_6])
    conv_6 = K.layers.Conv2D(512, 3, **conv)(merge_6)
    conv_6 = K.layers.Conv2D(512, 3, **conv)(conv_6)

    # Second decoder block
    up_7 = K.layers.UpSampling2D(size=(2, 2))(conv_6)
    up_7 = K.layers.Conv2D(256, 2, **conv)(up_7)

    # Concatenation of second decoder and third encoder block
    merge_7 = K.layers.Concatenate()([conv_3, up_7])
    conv_7 = K.layers.Conv2D(256, 3, **conv)(merge_7)
    conv_7 = K.layers.Conv2D(256, 3, **conv)(conv_7)

    # Third decoder block
    up_8 = K.layers.UpSampling2D(size=(2, 2))(conv_7)
    up_8 = K.layers.Conv2D(128, 2, **conv)(up_8)

    # Concatenation of third decoder and second encoder block
    merge_8 = K.layers.Concatenate()([conv_2, up_8])
    conv_8 = K.layers.Conv2D(128, 3, **conv)(merge_8)
    conv_8 = K.layers.Conv2D(128, 3, **conv)(conv_8)

    # Fourth decoder block
    up_9 = K.layers.UpSampling2D(size=(2, 2))(conv_8)
    up_9 = K.layers.Conv2D(64, 2, **conv)(up_9)

    # Concatenation of fourth decoder and first encoder block
    merge_9 = K.layers.Concatenate()([conv_1, up_9])
    conv_9 = K.layers.Conv2D(64, 3, **conv)(merge_9)
    conv_9 = K.layers.Conv2D(64, 3, **conv)(conv_9)

    # Output of U-Net
    conv_9 = K.layers.Conv2D(2, 3, **conv)(conv_9)
    conv_10 = K.layers.Conv2D(1, 1, activation="sigmoid")(conv_9)

    model = K.models.Model(inputs=[inputs], outputs=[conv_10], name="unet_simple")

    return model


def _encoder(x, conv, dropout=None, batch_norm=False):
    """
    Create an encoder block with dropout, batch normalization, kernel regularatization and
    more. It basically supports every possible parameter from the Keras API.

    Args:
        x (keras.Layer): Layer to build upon.
        conv (dict): All possible arguments from the Keras specification.
        dropout (float): Value for the dropout layer, between 0 and 1.
        batc_norm (bool): Whether to add batch normalization or not.

    Returns:
        some_layer (keras.Layer): Output layer with convolutions and additional
            parameters and layers.
    """
    if batch_norm:
        if dropout is not None:
            warn("Is it NOT recommended to use Batch Normalization with Dropout")

        conv["use_bias"] = False

        some_layer = K.layers.Conv2D(**conv)(x)
        some_layer = K.layers.Conv2D(**conv)(some_layer)
        some_layer = K.layers.BatchNormalization()

    else:
        some_layer = K.layers.Conv2D(**conv)(x)
        some_layer = K.layers.Conv2D(**conv)(some_layer)

    if dropout is not None:
        if dropout <= 0.0:
            raise ValueError("Dropout must be larger than zero.")
        some_layer = K.layers.Dropout(dropout)(some_layer)

    return some_layer


def _concatenate_and_upsample(prev_layer, cat_layer, upsample, conv):
    """
    Decoder blocks for the Unet where upsampling and concatenation take part.

    Args:
        prev_layer (keras.Layer): Previous layer from the network.
        cat_layer (keras.Layer): The concatenation layer that will be merged together
            with the upsampling one.
        upsample (Tuple[int, int]): Upsampling factor for rows and columns.
        conv (dict): All possible arguments from the Keras specification.

    Returns:
        output (keras.Layer): Result from the upsampling and concatenation.
    """
    some_layer = K.layers.UpSampling2D(size=upsample)(prev_layer)
    conv["kernel_size"] = 2
    some_layer = K.layers.Conv2D(**conv)(some_layer)

    merge_layer = K.layers.Concatenate()([cat_layer, some_layer])
    conv["kernel_size"] = 3
    output = K.layers.Conv2D(**conv)(merge_layer)
    output = K.layers.Conv2D(**conv)(output)

    return output


def custom_unet(
    input_size, conv, pool=None, dropout=None, batch_norm=None, up_sample=None
):
    """
    Implementation of the U-Net model but with extreme flexibility to add new paramaters
    like dropout, batch normalization, kernel regularizers and so on.
    
    Args:
        input_size (Tuple[int, int, int]): (height, width, channels), i.e.
        The information of the image.
        Must always be a multiple of 32, e.g. 256, 512.
        conv (dict): Keyword parameters from the Keras specification. This dictionary
            will affect ALL convolutional layers in the network.
        pool (Tuple[int, int]): Pooling windows
        dropout (float): Value for the dropout layer, between 0 and 1.
        batch_norm (bool): Add batch normalization to every encoder block.
        up_sample (Tuple[int, int]): Upsampling factor for rows and columns.
    Returns:
        model: A tf.keras.Model instance.
    """
    inputs = K.layers.Input(input_size)
    encoder_block = inputs
    encoding_layers = []

    for i in range(4):
        encoder_block = _encoder(
            encoder_block, conv, dropout=dropout, batch_norm=batch_norm
        )
        #
        encoding_layers.append(encoder_block)
        # Add the necessary pooling
        if pool is None:
            # Default pooling
            pool = (2, 2)
        encoder_block = K.layers.MaxPooling2D(pool_size=pool)(encoder_block)
        # Update filter size
        conv["filters"] *= 2

    # Add encoder-decoder block
    output_layer = _encoder(encoder_block, conv, dropout=dropout, batch_norm=batch_norm)

    for layer in reversed(encoding_layers):
        # Reduce number of filters
        conv["filters"] //= 2
        if up_sample is None:
            up_sample = (2, 2)
        output_layer = _concatenate_and_upsample(output_layer, layer, up_sample, conv)

    # Output of U-Net
    output_layer = K.layers.Conv2D(
        2, 3, activation=conv["activation"], padding=conv["padding"]
    )(output_layer)
    output_layer = K.layers.Conv2D(1, 1, activation="sigmoid")(output_layer)

    model = K.models.Model(inputs=[inputs], outputs=[output_layer], name="unet_custom")

    return model

